==> geiger_localstorage-0.6.17.tar.gz: lib/src/db/data/timestamp.dart <==
library geiger_localstorage;

import 'dart:async';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:logging/logging.dart';

class ExtendedTimestamp {
  static final Logger _logger = Logger('GeigerStorage');

  static const String timestampName = ':configTimestamp';

  /// offers the last know timestamp
  static int _lastTimestamp = 0;

  static int _lastWrite = 0;

  static bool _pendingWrite = true;

  static int _init = 0;

  static GenericController? _controller;

  static String? _uuid;

  static void reset() {
    _pendingWrite = false;
    _init = 0;
    _uuid = null;
    _controller = null;
    _lastTimestamp = -1;
    _lastWrite = -1;
  }

  static int now([bool requireInit = true]) {
    if (_init != 2 && requireInit) {
      throw StorageException('Extended timestamp not yet initialized');
    }
    int ret = DateTime.now().millisecondsSinceEpoch;
    if (ret <= _lastTimestamp) {
      ret = _lastTimestamp + 1;
    }
    _lastTimestamp = ret;
    if (requireInit) {
      _writeTimestampValueMicrotask();
    }
    _logger.log(Level.FINEST, 'Obtained timestamp $ret');
    return ret;
  }

  static String extendedNow([bool requireInit = true]) {
    return convertToExtendedTimestamp(now(requireInit));
  }

  static bool get init => _init == 2;

  static Future<void> _writeTimestampValueMicrotask() async {
    // make sure that:
    // - init has run
    // - last write was more than 10s away or no save has been done yet
    // - no write microtask is currently pending
    if (_init == 2 &&
        _lastTimestamp + 10000 > _lastWrite &&
        _lastTimestamp > -1 &&
        !_pendingWrite) {
      _lastWrite = _lastTimestamp;
      _pendingWrite = true;
      scheduleMicrotask(() async {
        // make sure that the node exists and has not been zapped in the meantime
        Node n = NodeImpl(timestampName, 'internal', null, Visibility.black);
        _lastTimestamp = now(false);
        await n.addValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        await _controller!.addOrUpdate(n);
        // write updated value
        await n.updateValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        await _controller!.update(n);
        _pendingWrite = false;
      });
    }
  }

  /// Initializes the timestamp with the persisted and possibly un-persisted changes of the database.
  static Future<void> initializeTimestamp(StorageController controller) async {
    if (_init == 0) {
      _init = 1;
      _logger.log(Level.FINE, 'Initializing timestamp');
      _controller = controller as GenericController;
      Node n;
      try {
        n = await _controller!.get(timestampName);
        if ((await n.getValue('lastUpdated')) == null) {
          _lastTimestamp = DateTime.now().millisecondsSinceEpoch;
          await n.addValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        }
      } on StorageException {
        n = NodeImpl(timestampName, 'internal', null, Visibility.black);
        _lastTimestamp = DateTime.now().millisecondsSinceEpoch;
        await n.addValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        try {
          await _controller!.add(n);
        } on StorageException {
          // someone else already added it... lets fetch it
          Node n2 = await _controller!.get(timestampName);
          if ((await n.getValue('lastUpdated')) != null) {
            n = n2;
          }
        }
      }
      String? lastUpdated = (await n.getValue('lastUpdated'))!.toSimpleString();

      _lastTimestamp = int.parse(lastUpdated!);
      _init = 2;

      List<Node> nl = await (_controller as GenericController).search(
          SearchCriteria(searchPath: ':', nodeValueLastModified: lastUpdated));
      for (Node n in nl) {
        if (_lastTimestamp < n.lastModified) {
          _lastTimestamp = n.lastModified;
        }
      }
      _logger.log(Level.INFO, 'timestamp initialized');
    } else if (_init == 1) {
      await _waitForInit();
    }
    if (_uuid == null) {
      try {
        Node n2 = await _controller!.get(':Local');
        _uuid = (await n2.getValue('currentDevice'))!.toSimpleString();
      } on StorageException {
        _logger.log(Level.SEVERE,
            'OOPS! No current device available in storage... contact developer');
      }
    }
  }

  static Future<void> _waitForInit() async {
    while (await Future<bool>.delayed(
        const Duration(milliseconds: 100), () => _init == 2)) {}
  }

  static String convertToExtendedTimestamp(int timestamp, [String? uuid]) {
    if (uuid != null) {
      _uuid = uuid;
    } else {
      uuid ??= _uuid;
    }
    return '${timestamp.toString().padLeft(20, '0')}-$_uuid';
  }

  static int convertFromExtendedTimestamp(String extendedTimestamp) {
    String regexString =
        r'^0*([0-9]+)-([0-9a-f]{8,8}-[0-9a-f]{4,4}-[0-9a-f]{4,4}-[0-9a-f]{4,4}-[0-9a-f]{12,12})$';
    RegExp regExp = RegExp(regexString);
    Iterable<RegExpMatch> matches = regExp.allMatches(extendedTimestamp);
    if (matches.length != 1) {
      throw StorageException('i'
          'nvalid timestamp "$extendedTimestamp"');
    }
    RegExpMatch m = matches.elementAt(0);
    return int.parse(m.group(1)!);
  }
}

==> geiger_localstorage-0.6.21.tar.gz: lib/src/db/data/timestamp.dart <==
library geiger_localstorage;

import 'dart:async';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:logging/logging.dart';

class ExtendedTimestamp {
  static final Logger _logger = Logger('GeigerStorage');
  static const String _regexExtendedTimestampString =
      r'^0*([0-9]+)-([0-9a-f]{8,8}-[0-9a-f]{4,4}-[0-9a-f]{4,4}-[0-9a-f]{4,4}-[0-9a-f]{12,12})$';

  static const String timestampName = ':configTimestamp';

  /// offers the last know timestamp
  static int _lastTimestamp = 0;

  static int _lastWrite = 0;

  static bool _pendingWrite = true;

  static int _init = 0;

  static GenericController? _controller;

  static String? _uuid;

  static void reset() {
    _pendingWrite = false;
    _init = 0;
    _uuid = null;
    _controller = null;
    _lastTimestamp = -1;
    _lastWrite = -1;
  }

  static int now([bool requireInit = true]) {
    if (_init != 2 && requireInit) {
      throw StorageException('Extended timestamp not yet initialized');
    }
    int ret = DateTime.now().millisecondsSinceEpoch;
    if (ret <= _lastTimestamp) {
      ret = _lastTimestamp + 1;
    }
    _lastTimestamp = ret;
    if (requireInit) {
      _writeTimestampValueMicrotask();
    }
    _logger.log(Level.FINEST, 'Obtained timestamp $ret');
    return ret;
  }

  static String extendedNow([bool requireInit = true]) {
    return convertToExtendedTimestamp(now(requireInit));
  }

  static bool get init => _init == 2;

  static Future<void> _writeTimestampValueMicrotask() async {
    // make sure that:
    // - init has run
    // - last write was more than 10s away or no save has been done yet
    // - no write microtask is currently pending
    if (_init == 2 &&
        _lastTimestamp + 10000 > _lastWrite &&
        _lastTimestamp > -1 &&
        !_pendingWrite) {
      _lastWrite = _lastTimestamp;
      _pendingWrite = true;
      scheduleMicrotask(() async {
        // make sure that the node exists and has not been zapped in the meantime
        Node n = NodeImpl(timestampName, 'internal', null, Visibility.black);
        _lastTimestamp = now(false);
        await n.addValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        await _controller!.addOrUpdate(n);
        // write updated value
        await n.updateValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        await _controller!.update(n);
        _pendingWrite = false;
      });
    }
  }

  /// Initializes the timestamp with the persisted and possibly un-persisted changes of the database.
  static Future<void> initializeTimestamp(StorageController controller) async {
    if (_init == 0) {
      _init = 1;
      _logger.log(Level.FINE, 'Initializing timestamp');
      _controller = controller as GenericController;
      Node n;
      try {
        n = await _controller!.get(timestampName);
        if ((await n.getValue('lastUpdated')) == null) {
          _lastTimestamp = DateTime.now().millisecondsSinceEpoch;
          await n.addValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        }
      } on StorageException {
        n = NodeImpl(timestampName, 'internal', null, Visibility.black);
        _lastTimestamp = DateTime.now().millisecondsSinceEpoch;
        await n.addValue(NodeValueImpl('lastUpdated', '$_lastTimestamp'));
        try {
          await _controller!.add(n);
        } on StorageException {
          // someone else already added it... lets fetch it
          Node n2 = await _controller!.get(timestampName);
          if ((await n.getValue('lastUpdated')) != null) {
            n = n2;
          }
        }
      }
      String? lastUpdated = (await n.getValue('lastUpdated'))!.toSimpleString();

      _lastTimestamp = int.parse(lastUpdated!);
      _init = 2;

      List<Node> nl = await (_controller as GenericController).search(
          SearchCriteria(searchPath: ':', nodeValueLastModified: lastUpdated));
      for (Node n in nl) {
        if (_lastTimestamp < n.lastModified) {
          _lastTimestamp = n.lastModified;
        }
      }
      _logger.log(Level.INFO, 'timestamp initialized');
    } else if (_init == 1) {
      await _waitForInit();
    }
    if (_uuid == null) {
      try {
        Node n2 = await _controller!.get(':Local');
        _uuid = (await n2.getValue('currentDevice'))!.toSimpleString();
      } on StorageException {
        _logger.log(Level.SEVERE,
            'OOPS! No current device available in storage... contact developer');
      }
    }
  }

  static Future<void> _waitForInit() async {
    while (await Future<bool>.delayed(
        const Duration(milliseconds: 100), () => _init == 2)) {}
  }

  static String convertToExtendedTimestamp(int timestamp, [String? uuid]) {
    if (uuid != null) {
      _uuid = uuid;
    } else {
      uuid ??= _uuid;
    }
    return '${timestamp.toString().padLeft(20, '0')}-$_uuid';
  }

  /// Accepts either a timestamp or an extended timestamp as input and returns the timestamp value only.
  static int convertFromExtendedTimestamp(String extendedTimestamp) {
    if (!isExtendedTimestamp(extendedTimestamp)) {
      throw StorageException('invalid timestamp "$extendedTimestamp"');
    }
    RegExp regExp = RegExp(_regexExtendedTimestampString);
    Iterable<RegExpMatch> matches = regExp.allMatches(extendedTimestamp);
    RegExpMatch m = matches.elementAt(0);
    return int.parse(m.group(1)!);
  }

  /// Tests if a timestamp is a valid extended timestamp
  static bool isExtendedTimestamp(String extendedTimestamp) {
    RegExp regExp = RegExp(_regexExtendedTimestampString);
    Iterable<RegExpMatch> matches = regExp.allMatches(extendedTimestamp);
    return matches.length == 1;
  }
}
